# Gridlock - a collection of grid-based strategy games
# Copyright (c) 2009 Brian Nenninger (brian@dozingcatsoftware.com)
# Gridlock comes with ABSOLUTELY NO WARRANTY, and is released under version 3 of the GNU GPL.

def import_numpy():
    """ Returns the numpy module if available, otherwise fake_numpy """
    try:
        return __import__('numpy')
    except:
        return __import__('fake_numpy')
        
import itertools
import collections

numpy = import_numpy()

def create_grid(dim):
    return numpy.empty(dim, int)

# add/subtract/negate optimized for 2d and 3d cases
def add_tuples(a,b):
    if len(a)==2: return (a[0]+b[0],a[1]+b[1])
    if len(a)==3: return (a[0]+b[0],a[1]+b[1],a[2]+b[2])
    return tuple(x+y for x,y in itertools.izip(a,b))
    
def subtract_tuples(a,b):
    if len(a)==2: return (a[0]-b[0],a[1]-b[1])
    if len(a)==3: return (a[0]-b[0],a[1]-b[1],a[2]-b[2])
    return tuple(x-y for x,y in itertools.izip(a,b))
    
def negate_tuple(a):
    if len(a)==2: return (-a[0],-a[1])
    if len(a)==3: return (-a[0],-a[1],-a[2])
    return tuple(-x for x in a)
    
def max_coord_diff(a,b):
    if len(a)==2: return max(abs(a[0]-b[0]), abs(a[1]-b[1]))
    return max(abs(x-y) for x,y in itertools.izip(a,b))

index_array_cache = {}

def index_array(dim):
    global index_array_cache
    a = index_array_cache.get(dim)
    if a is None:
        # create array where the value at each tuple index is the index itself
        a = numpy.empty(dim, object)
        for c in numpy.ndindex(dim):
            a[c] = c
        index_array_cache[dim] = a
    return a
    
direction_array_cache = {}

def direction_array(ndim):
    global direction_array_cache
    a = direction_array_cache.get(ndim)
    if a is None:
        # get indices for n-dimensional array of size 3, subtract 1 to go from -1 to +1
        a = []
        sub1 = (-1,)*ndim
        for c in numpy.ndindex(tuple([3]*ndim)):
            a.append(add_tuples(c,sub1))
        # remove "direction" with all zeros
        a.remove(tuple([0]*ndim))
        direction_array_cache[ndim] = a
    return a

ortho_direction_cache = {}

def ortho_direction_array(ndim):
    global ortho_direction_cache
    a = ortho_direction_cache.get(ndim)
    if a is None:
        a = []
        for i in xrange(ndim):
            d1 = [0]*ndim
            d1[i] = -1
            d2 = d1[:]
            d2[i] = +1
            a.extend([d1,d2])
        ortho_direction_cache[ndim] = a
    return a
            

def neighbor_slice(arr, center, dist=1):
    slices = tuple([slice(max(c-dist,0), c+dist+1) for c in center])
    return arr[slices]

def valid_coords(grid, c):
    d = len(c)
    if d!=grid.ndim: return False
    if d==2: return 0<=c[0]<grid.shape[0] and 0<=c[1]<grid.shape[1]
    if d==3: return 0<=c[0]<grid.shape[0] and 0<=c[1]<grid.shape[1] and 0<=c[2]<grid.shape[2]
    return all(0<=cval<grid.shape[i] for i,cval in enumerate(c))

def itervalues(grid):
    return grid.flat
    
def itercoords(grid):
    return numpy.ndindex(grid.shape)
    
def enumerate_all(grid):
    return numpy.ndenumerate(grid)
    
def neighbor_values(grid, center, dist=1):
    return neighbor_slice(grid, center, dist)
    
def neighbor_coords(grid, center, dist=1):
    n = neighbor_slice(index_array(grid.shape), center, dist)
    return n.flat

def enumerate_neighbors(grid, center, dist=1):
    vals = neighbor_slice(grid, center, dist)
    coords = neighbor_slice(index_array(grid.shape), center, dist)
    return itertools.izip(coords.flat, vals.flat)
    
def enumerate_ortho_neighbors(grid, center):
    for d in ortho_direction_array(grid.ndim):
        p = add_tuples(center, d)
        if valid_coords(grid, p):
            yield (p, grid[p])
    
def enumerate_coords_with_value(grid, val):
    for c, v in numpy.ndenumerate(grid):
        if v==val: yield c
        
def first_coord_with_value(grid, val):
    for c, v in numpy.ndenumerate(grid):
        if v==val: return c
    return None
        
def enumerate_neighbor_coords_with_value(grid, center, val, dist=1):
    for c,v in enumerate_neighbors(grid, center, dist):
        if v==val: yield c
        
def enumerate_values_with_direction(grid, start, direction):
    done = False
    c = start
    while not done:
        c = add_tuples(c, direction)
        if valid_coords(grid, c):
            yield c, grid[c]
        else:
            done = True
            
def connected_positions(grid, pos, predicate, ortho=True):
    """ Returns all positions reachable starting from pos and moving orthoganally (or diagonally if ortho is False)
        through positions for which predicate(position, value) is true. 
    """
    neighbor_func = enumerate_ortho_neighbors if ortho else enumerate_neighbors
    # use a queue starting with pos and pushing valid neighbors as we find them
    queue = collections.deque((pos,))
    seen = set((pos,))
    while queue:
        p = queue.popleft()
        for np,nval in neighbor_func(grid, p):
            if np not in seen and predicate(np, nval):
                # add immediately to seen set so we don't push any copies on the queue
                queue.append(np)
                seen.add(np)
    return seen

def path_between_positions(grid, from_pos, to_pos, predicate, ortho=True):
    """ Returns a list of positions of orthoganal (or diagonal if ortho is False) moves starting from from_pos
        and ending at to_pos, where predicate(position, value) is true for each intermediate position.
    """
    neighbor_func = enumerate_ortho_neighbors if ortho else enumerate_neighbors
    # queue stores (position, path_list) tuples
    queue = collections.deque()
    queue.append([from_pos, [from_pos]])
    seen = set((from_pos,))
    # mostly the same algorithm as connected_positions, keep track of seen positions and their paths
    while queue:
        p,path = queue.popleft()
        for np,nval in neighbor_func(grid, p):
            if np==to_pos:
                # found goal
                return path+[to_pos]
            elif np not in seen and predicate(np, nval):
                queue.append((np, path+[np]))
                seen.add(np)
    # no path to to_pos
    return None
    
def connected_groups_with_predicate(grid, predicate, ortho=True):
    """ Yields sets of connected positions matching the predicate (2-arg function taking position and value) """
    used = set()
    for p,val in enumerate_all(grid):
        if predicate(p, val) and p not in used:
            # get connections from this position with same value
            current_set = connected_positions(grid, p, predicate, ortho)
            used.update(current_set)
            yield current_set

def connected_groups(grid, val, ortho=True):
    """ Yields sets of connected positions with given value """
    return connected_groups_with_predicate(grid, lambda gpos,gval: gval==val, ortho)
            